/*******************************************************************************

Copyright (c) 2003, Robert Cowham and Vaccaperna Systems Ltd.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL VACCAPERNA SYSTEMS LTD. BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

/*******************************************************************************
 * Name:	p4.cpp
 *
 * Author:	Robert Cowham <robert@vaccaperna.co.uk>
 *
 * Description: 
 *		Implementation of p4 - the COM class and thus external interface.
 *
 ******************************************************************************/

#include "stdafx.h"
#include "P4com.h"
#include "p4.h"

#include "TraceUtils.h"

#define CheckInitialized() \
	if (!m_Initialized) return Error("p4 init method not called!");

/////////////////////////////////////////////////////////////////////////////
// p4

STDMETHODIMP p4::InterfaceSupportsErrorInfo(REFIID riid)
{
	if (riid == IID_Ip4)
		return S_OK;
	return E_FAIL;
}


STDMETHODIMP p4::get_Charset(BSTR *pVal)
{
	*pVal = m_client.GetCharset();
	return S_OK;
}

STDMETHODIMP p4::put_Charset(BSTR newVal)
{
	m_client.SetCharset(newVal);
	return S_OK;
}

STDMETHODIMP p4::get_Client(BSTR *pVal)
{
	*pVal = m_client.GetClient();
	return S_OK;
}

STDMETHODIMP p4::put_Client(BSTR newVal)
{
	m_client.SetClient(newVal);
	return S_OK;
}

STDMETHODIMP p4::get_Cwd(BSTR *pVal)
{
	*pVal = m_client.GetCwd();
	return S_OK;
}

STDMETHODIMP p4::put_Cwd(BSTR newVal)
{
	m_client.SetCwd(newVal);
	return S_OK;
}

STDMETHODIMP p4::get_Host(BSTR *pVal)
{
	*pVal = m_client.GetHost();
	return S_OK;
}

STDMETHODIMP p4::put_Host(BSTR newVal)
{
	m_client.SetHost(newVal);
	return S_OK;
}

STDMETHODIMP p4::get_Language(BSTR *pVal)
{
	*pVal = m_client.GetLanguage();
	return S_OK;
}

STDMETHODIMP p4::put_Language(BSTR newVal)
{
	m_client.SetLanguage(newVal);
	return S_OK;
}

STDMETHODIMP p4::get_Os(BSTR *pVal)
{
	*pVal = m_client.GetOs();
	return S_OK;
}

STDMETHODIMP p4::get_Password(BSTR *pVal)
{
	*pVal = m_client.GetPassword();
	return S_OK;
}

STDMETHODIMP p4::put_Password(BSTR newVal)
{
	m_client.SetPassword(newVal);
	return S_OK;
}

STDMETHODIMP p4::get_Port(BSTR *pVal)
{
	*pVal = m_client.GetPort();
	return S_OK;
}

STDMETHODIMP p4::put_Port(BSTR newVal)
{
	m_client.SetPort(newVal);
	return S_OK;
}


STDMETHODIMP p4::get_User(BSTR *pVal)
{
	*pVal = m_client.GetUser();
	return S_OK;
}

STDMETHODIMP p4::put_User(BSTR newVal)
{
	m_client.SetUser(newVal);
	return S_OK;
}

STDMETHODIMP p4::run(BSTR cmd, SAFEARRAY **valArray)
{
	String errMsg; 
	Long result = m_client.Run(cmd, errMsg);
	ArrayString *as;
	as = (ArrayString *)valArray;

	(*as) = m_client.m_ui.InfoArray();
	if (result)
	{
		errMsg = m_client.ErrorMsg();
		return Error((LPCSTR)errMsg, GUID_NULL, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_NULL, 1));
	}
	else
	{
		errMsg = m_client.ErrorMsg();
		if (!errMsg.IsNull() && !errMsg.IsEmpty())
		{
			return Error((LPCSTR)errMsg, GUID_NULL, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_NULL, 2));
		}
	}
	return S_OK;
}

HRESULT ConvertArrayStringToVariant(ArrayString &as, VARIANT *pvaVariant)
{
	// Use technique shown in:
	// http://support.microsoft.com/default.aspx?scid=http://support.microsoft.com:80/support/kb/articles/Q165/9/67.ASP&NoWebContent=1
    HRESULT hr = NOERROR;
    LPSAFEARRAY psa;
    SAFEARRAYBOUND rgsabound[]  = {as.Elements(), 0 }; // 0-based
    int i;

    if (!pvaVariant)
    {
	    return E_INVALIDARG;
    }
    VariantInit(pvaVariant);
    psa = SafeArrayCreate(VT_VARIANT, 1, rgsabound);
    if (!psa)
    {
		return E_OUTOFMEMORY;
    }

    VARIANT *va = new VARIANT[as.Elements()];
	if (!va)
	{
		hr = E_OUTOFMEMORY;
		goto Error_Block;
	}
    for (i = 0; i < as.Elements(); i++)
    {
		VariantInit(&va[i]);
		V_VT(&va[i]) = VT_BSTR;
	    V_BSTR(&va[i]) = SysAllocString(as[i]);
	    if (!V_BSTR(&va[i]))
		{
			hr = E_OUTOFMEMORY;
			goto Error_Block;
		}
    }

    {
        //Plug references to the data into the SAFEARRAY
        LPVARIANT rgElems;
        if (FAILED(hr = SafeArrayAccessData(psa,(LPVOID*)&rgElems)))
	    {
	        goto Error_Block;
		}
		for (i = 0; i < as.Elements(); i++)
		{
			rgElems[i] = va[i];
		}
		SafeArrayUnaccessData(psa);
    }

    V_VT(pvaVariant) = VT_ARRAY | VT_VARIANT;
    V_ARRAY(pvaVariant) = psa;

    return S_OK;

Error_Block:
    for (i = 0; i < as.Elements(); i++)
    {
		if (V_BSTR(&va[i]))
		{
			VariantClear(&va[i]);
		}
    }
    return hr;
}

STDMETHODIMP p4::run_variant(BSTR cmd, VARIANT *pvaVariant)
{
	String errMsg; 
	Long result = m_client.Run(cmd, errMsg);
	ArrayString as = m_client.m_ui.InfoArray();
	HRESULT hr = ConvertArrayStringToVariant(as, pvaVariant);
	if (hr)
		return hr;
	if (result)
	{
		errMsg = m_client.ErrorMsg();
		return Error((LPCSTR)errMsg, GUID_NULL, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_NULL, 1));
	}
	else
	{
		errMsg = m_client.ErrorMsg();
		if (!errMsg.IsNull() && !errMsg.IsEmpty())
		{
			return Error((LPCSTR)errMsg, GUID_NULL, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_NULL, 2));
		}
	}
	return S_OK;
}

STDMETHODIMP p4::get_Errors(SAFEARRAY **valArray)
{
	ArrayString *as;
	as = (ArrayString *)valArray;

	(*as) = m_client.m_ui.ErrorArray();
	return S_OK;
}

STDMETHODIMP p4::get_Errors_variant(VARIANT *vArray)
{
	ArrayString as = m_client.m_ui.ErrorArray();
	return ConvertArrayStringToVariant(as, vArray);
}

STDMETHODIMP p4::get_Warnings(SAFEARRAY **valArray)
{
	ArrayString *as;
	as = (ArrayString *)valArray;

	(*as) = m_client.m_ui.WarningArray();
	return S_OK;
}

STDMETHODIMP p4::get_Warnings_variant(VARIANT *vArray)
{
	ArrayString as = m_client.m_ui.WarningArray();
	return ConvertArrayStringToVariant(as, vArray);
}

STDMETHODIMP p4::Connect()
{
	String errMsg; 
	Long result = m_client.Connect(errMsg);
	if (result)
		return Error((LPCSTR)errMsg);
	return S_OK;
}

STDMETHODIMP p4::Disconnect()
{
	m_client.Disconnect();
	return S_OK;
}

STDMETHODIMP p4::get_ExceptionLevel(long *pVal)
{
	*pVal = m_client.ExceptionLevel();
	return S_OK;
}

STDMETHODIMP p4::put_ExceptionLevel(long newVal)
{
	m_client.ExceptionLevel(newVal);
	return S_OK;
}

STDMETHODIMP p4::Tagged()
{
	m_client.Tagged();
	return S_OK;
}

STDMETHODIMP p4::ParseForms()
{
	m_client.ParseForms();
	return S_OK;
}

STDMETHODIMP p4::get_TempFilename(BSTR *pVal)
{
	String f;
	f = m_client.m_ui.BinaryFileName();
	if (f.IsNull() || f.IsEmpty())
		f = m_client.m_ui.TextFileName();

	*pVal = f;
	return S_OK;
}

STDMETHODIMP p4::get_Var(BSTR varName, BSTR *pVal)
{
	String val = m_client.m_ui.GetVar(varName);
	if (val.IsNull())
	{
		val = "";
		*pVal = val;
		if (m_client.ExceptionLevel())
			return Error("Var doesn't exist!");
	}
	*pVal = val;
	return S_OK;
}

STDMETHODIMP p4::put_Var(BSTR varName, BSTR newVal)
{
	m_client.m_ui.SetVar(varName, newVal);
	return S_OK;
}

STDMETHODIMP p4::get_VarExists(BSTR varName, BOOL *pVal)
{
	String val = m_client.m_ui.GetVar(varName);
	if (val.IsNull())
	{
		*pVal = False;
	}
	else
	{
		*pVal = True;
	}
	return S_OK;
}

STDMETHODIMP p4::get_ArrayVar(BSTR varName, SAFEARRAY **pVal)
{
	ArrayString *as = (ArrayString *)pVal;
	(*as) = m_client.m_ui.GetArrayVar(varName);

	return S_OK;
}

STDMETHODIMP p4::get_ArrayVar_variant(BSTR varName, VARIANT *vArray)
{
	ArrayString as = m_client.m_ui.GetArrayVar(varName);
	return ConvertArrayStringToVariant(as, vArray);
}

STDMETHODIMP p4::put_ArrayVar(BSTR varName, SAFEARRAY **newVal)
{
	ArrayString *as = CheckEmptyArray(newVal);
	m_client.m_ui.SetArrayVar(varName, *as);

	return S_OK;
}

STDMETHODIMP p4::put_ArrayVar_variant(BSTR varName, SAFEARRAY **newVal)
{
	ArrayVariant *av = (ArrayVariant *)newVal;
	ArrayString as(Dim(av->Elements()));
	String s;
	Variant v;
	for (int i = 0; i < av->Elements(); i++)
	{
		v = (*av)[i];
		s = (BSTR)v;
		as[i] = s;
	}
	m_client.m_ui.SetArrayVar(varName, as);

	return S_OK;
}

ArrayString* p4::CheckEmptyArray(SAFEARRAY **newVal)
{

	if (*newVal == NULL)
	{
		return &m_emptyArray;
	}
	else
	{
		m_array = *(ArrayString *)newVal;
		return &m_array;
	}
}

String p4::CheckEmptyString(BSTR newVal)
{
	if (newVal == NULL)
	{
		return m_emptyString;
	}
	String s(newVal);
	return s;
}


STDMETHODIMP p4::get_ServerVersion(long *pVal)
{
	*pVal = m_client.GetServerVersion();
	return S_OK;
}
